<?php

namespace Drupal\preview_graphql\Services;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\graphql\Utility\StringHelper;
use Drupal\preview_graphql\Entity\PreviewGraphQL;
use GraphQL\Server\OperationParams;
use Drupal\graphql\GraphQL\Execution\QueryProcessor;

/**
 * Class PreviewGraphQLManager
 *
 * @package Drupal\preview_graphql\Service
 */
class PreviewGraphQLManager {

  /**
   * A entity type manager instance.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The preview graphql global settings.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The preview graphql backend service.
   *
   * @var \Drupal\preview_graphql\Services\PreviewGraphQLBackend
   */
  protected $previewGraphQlBackend;


  /**
   * The graphql query processor service.
   *
   * @var \Drupal\graphql\GraphQL\Execution\QueryProcessor
   */
  protected $queryProcessor;

  protected $dateTime;

  const PREFIX_CACHE = 'pgql_';

  // Max 24h.
  const DEFAULT_TIME_CACHE = 86400;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ConfigFactoryInterface $config_factory, PreviewGraphQLBackend $preview_graphql_backend, QueryProcessor $query_processor, TimeInterface $date_time) {
    $this->entityTypeManager = $entity_type_manager;
    $this->config = $config_factory->get('preview_graphql.settings');
    $this->previewGraphQlBackend = $preview_graphql_backend;
    $this->queryProcessor = $query_processor;
    $this->dateTime = $date_time;
  }

  /**
   * Return the preview graphql backend service.
   *
   * @return \Drupal\preview_graphql\Services\PreviewGraphQLBackend
   *   The preview graphql backend service.
   */
  public function getPreviewGraphQLBackend() {
    return $this->previewGraphQlBackend;
  }

  /**
   * Get preview graphql entity by entity type and/or bundle.
   *
   * @param string $entity_type
   *   The entity type of preview graphql entity.
   * @param array $bundles
   *   The entity bundle of preview graphql entity.
   * @param bool $load
   *   If true, load all preview graphql entity.
   *
   * @return array|\Drupal\preview_graphql\Entity\PreviewGraphQL[]|int
   *   The ids of preview graph ql entity or loaded entities.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getPreviewGraphQlEntityByTypeBundle($entity_type, $bundle, $load = FALSE) {
    $query = $this->entityTypeManager
      ->getStorage('preview_graphql')->getQuery();

    $query->condition('entity_type', $entity_type, '=');
    //@todo change this part of code when core fixed.
    if (!empty($bundle)) {
      // @see https://www.drupal.org/project/drupal/issues/3053941
      //$query->condition('bundle', [$bundle], 'IN');
    }
    else {
      $query->condition('bundle', NULL, 'IS NULL');
    }
    $ids = $query->execute();
    if (!empty($ids) && (!empty($bundle) || $load = TRUE)) {

      $entities = $this->entityTypeManager
        ->getStorage('preview_graphql')->loadMultiple($ids);
      foreach ($entities as $entity_id => $preview_graphql_entity) {
        if (!in_array($bundle, $preview_graphql_entity->get('bundle'))) {
          unset($entities[$entity_id]);
        }
      }
      if ($load) {
        return $entities;
      }
      return array_keys($entities);
    }
    if (empty($ids) || $load == FALSE) {
      return $ids;
    }

  }

  /**
   * Build the front url to send jwt.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *
   * @return array|bool|mixed|string|null
   *
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getUrlFrontByEntity(EntityInterface $entity) {
    // Get url front.
    $url_front = $this->config->get('url_front');
    if (empty($url_front)) {
      $message = t('No front url set in global settings');
      $this->previewGraphQlBackend->getMessenger()->addError($message);
      $this->previewGraphQlBackend->getLogger()->error($message);
      return FALSE;
    }
    // Get the uri.
    $preview_graphql_entity = $this->getPreviewGraphQlEntity($entity);
    if (empty($preview_graphql_entity)) {
      $uri_front = $this->config->get('uri_front');
    }
    else {
      $uri_front = $this->getSetting($preview_graphql_entity, 'uri_front');
    }
    if (substr($uri_front, 0, 1) != '/') {
      $url_front .= '/';
    }
    $url_front .= $uri_front;
    return $url_front;
  }

  /**
   * Load all preview graphql entity and merge deriver type.
   *
   * @return array
   *   The merged of deriver type.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDeriverType() {

    $deriver_type = ['entity_type' => [], 'bundle' => []];
    $previews_graphqls = $this->entityTypeManager->getStorage('preview_graphql')
      ->loadMultiple();
    foreach ($previews_graphqls as $previews_graphql) {
      if ($previews_graphql->get('deriver') == 'bundle' && !empty($previews_graphql->get('bundle'))) {
        foreach ($previews_graphql->get('bundle') as $bundle) {
          $deriver_type['bundle'][$previews_graphql->get('entity_type')][$bundle] = $bundle;
        }
        continue;
      }
      $deriver_type['entity_type'][$previews_graphql->get('entity_type')] = $previews_graphql->get('entity_type');
    }
    return $deriver_type;
  }

  /**
   * Return the preview graphql entity corresponding of an drupal entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity for which we search preview graphql entity.
   *
   * @return bool|\Drupal\preview_graphql\Entity\PreviewGraphQL
   *   False ifno preview graphql entity found, else the entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getPreviewGraphQlEntity(EntityInterface $entity) {
    $preview_graphql_entity = $this->getPreviewGraphQlEntityByTypeBundle($entity->getEntityTypeId(), $entity->bundle(), TRUE);
    if (empty($preview_graphql_entity)) {
      $preview_graphql_entity = $this->getPreviewGraphQlEntityByTypeBundle($entity->getEntityTypeId(), '', TRUE);
    }
    return !empty($preview_graphql_entity) ? reset($preview_graphql_entity) : FALSE;
  }

  /**
   * Build the graphql query for entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to build query.
   *
   * @return bool|string
   *   False if error, else the graphql query.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function buildQueryByEntity(EntityInterface $entity) {
    $preview_graphql_entity = $this->getPreviewGraphQlEntity($entity);

    if (empty($preview_graphql_entity)) {
      $message = t('No preview graphql entity found for this entity.');
      $this->previewGraphQlBackend->getMessenger()->addError($message);
      $this->previewGraphQlBackend->getLogger()->error($message);
      return FALSE;
    }
    $deriver = $preview_graphql_entity->get('deriver');

    // Init query name depend what query expect.
    $query_name = StringHelper::propCase('Preview_' . $entity->getEntityTypeId() . '_' . $entity->bundle());
    if ($deriver == 'entity_type') {
      $query_name = StringHelper::propCase('Preview_' . $entity->getEntityTypeId());
    }
    // Insert args into query gql.
    $query_name .= '(entity_type:"' . $entity->getEntityTypeId() . '",uuid:"' . $entity->uuid() . '")';
    $query_gql = $this->getSetting($preview_graphql_entity, 'query_gql');
    $fragment_gql = $this->getSetting($preview_graphql_entity, 'fragment_gql');
    // Insert query name, query gql, and fragment.
    $query = '{' . $query_name . '{' . $query_gql . '}}' . $fragment_gql;
    return $query;
  }

  /**
   * Process the entity query into graphQL engine.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to process.
   *
   * @return bool|\Drupal\graphql\GraphQL\Execution\QueryResult
   *   False if error, else the query result of graph ql.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\TempStore\TempStoreException
   */
  public function processQueryEntity(EntityInterface $entity) {
    // Process the graphql query.
    $query = $this->buildQueryByEntity($entity);
    // @todo get it dynamically.
    $schema = 'default:default';
    $operation = OperationParams::create(['query' => $query]);
    /* @var  $result \Drupal\graphql\GraphQL\Execution\QueryResult */
    $result = $this->queryProcessor->processQuery($schema, $operation);
    // Graphql Errors.
    /* @var  $errors \GraphQL\Error\Error */
    foreach ($result->errors as $errors) {
      $message = $errors->getMessage();
      $this->previewGraphQlBackend->getMessenger()->addError($message);
      $this->previewGraphQlBackend->getLogger()->error($message);
      // Remove from temp store.
      $this->previewGraphQlBackend->deleteTempStore($entity->getEntityTypeId() . '_' . $entity->uuid());
      return FALSE;
    }
    return $result;
  }

  /**
   * Get settings of the preview graphql entity.
   *
   * Also return the default settings if not override.
   *
   * @param \Drupal\preview_graphql\Entity\PreviewGraphQL $preview_graph_ql
   *   The preview graphql entity.
   * @param string $key
   *   The config name.
   *
   * @return string|null
   *   The config value.
   */
  public function getSetting(PreviewGraphQL $preview_graph_ql, $key) {
    $override = $preview_graph_ql->get('override');
    if (!empty($override['enable']) && $override['enable'] == TRUE && !empty($override[$key])) {
      return $override[$key];
    }
    return $this->config->get($key);
  }

  /**
   * Return an data set into cache bin. I.E: form_state, gql,callback_url.
   *
   * @param string $cache_data_key
   *   The key searched into cache. I.E: form_state, gql,callback_url.
   *
   * @return array|string
   *   IF array, contains error (key 'error'), else return data searched.
   */
  public function getCacheDataValueByJwt($cache_data_key) {
    $cid = $this->previewGraphQlBackend->getCidByJwt();
    if (is_array($cid)) {
      return $cid;
    }
    $cache = $this->previewGraphQlBackend->getCacheBackend()->get($cid);
    if (empty($cache->data[$cache_data_key])) {
      return ['error' => t('Invalid jwt, %key is empty.', ['%key' => $cache_data_key])];
    }
    return $cache->data[$cache_data_key];
  }

  /**
   * Set into cache data.
   *
   * @param string $cid
   *   The cache id where to save data.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The data form_state of entity form.
   * @param array $graph_ql_resut
   *   The graphql data result.
   * @param string $url_callback
   *   The url of entity form.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The old entity.
   *
   * @return array|bool
   *   IF array, contains error (key 'error'), else return true.
   */
  public function setEntityCache($cid, FormStateInterface $form_state, array $graph_ql_resut, $url_callback, EntityInterface $entity) {
    if (empty($cid) || empty($form_state) || empty($graph_ql_resut) || empty($url_callback) || empty($entity)) {

      return ['error' => t("Can't set cache some data are empty.")];
    }
    $cache = [];
    if (!empty($entity->getCacheTags())) {
      $cache = $entity->getCacheTags();
    }
    // Add our cache tag.
    $cache[] = $this->getTagName($entity);
    $this->previewGraphQlBackend->getCacheBackend()->set($cid, [
      'gql' => $graph_ql_resut,
      'form_state' => $form_state,
      'callback_url' => $url_callback,
    ], $this->dateTime->getRequestTime() + self::DEFAULT_TIME_CACHE, $cache);
    return TRUE;
  }

  /**
   * Return the preview graphql tag name for an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to build the tag name.
   *
   * @return string
   *   The tag name.
   */
  public function getTagName(EntityInterface $entity) {
    return self::PREFIX_CACHE . $entity->getEntityTypeId() . ':' . $entity->uuid();;
  }

  /**
   * Delete the cache of entity.
   *
   * @param string|null $cid
   *   The cache id to delete.
   *
   * @return array|string|null
   *   IF array, contains error (key 'error').
   */
  public function deleteEntityCache($cid = NULL) {
    if (empty($cid)) {
      $cid = $this->previewGraphQlBackend->getCidByJwt();
      if (is_array($cid)) {
        return $cid;
      }
    }
    $this->previewGraphQlBackend->getCacheBackend()->delete($cid);
    return TRUE;
  }

}
